@@ -12,9 +12,10 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import os, platform, re, sys, shutil, subprocess, typing
+import os, platform, re, sys, shutil, subprocess
 import tempfile
 import shlex
+import typing as T
 
 from . import coredata
 from .linkers import ArLinker, ArmarLinker, VisualStudioLinker, DLinker, CcrxLinker, IntelVisualStudioLinker
@@ -46,7 +47,8 @@ from .linkers import (
     CcrxDynamicLinker,
     ClangClDynamicLinker,
     DynamicLinker,
-    GnuDynamicLinker,
+    GnuBFDDynamicLinker,
+    GnuGoldDynamicLinker,
     LLVMDynamicLinker,
     MSVCDynamicLinker,
     OptlinkDynamicLinker,
@@ -110,7 +112,7 @@ from .compilers import (
 
 build_filename = 'meson.build'
 
-CompilersDict = typing.Dict[str, Compiler]
+CompilersDict = T.Dict[str, Compiler]
 
 def detect_gcovr(min_version='3.3', new_rootdir_version='4.2', log=False):
     gcovr_exe = 'gcovr'
@@ -167,7 +169,7 @@ def detect_ninja_command_and_version(version: str = '1.5', log: bool = False) ->
                 mlog.log('Found {}-{} at {}'.format(name, found, quote_arg(n)))
             return (n, found)
 
-def get_llvm_tool_names(tool: str) -> typing.List[str]:
+def get_llvm_tool_names(tool: str) -> T.List[str]:
     # Ordered list of possible suffixes of LLVM executables to try. Start with
     # base, then try newest back to oldest (3.5 is arbitrary), and finally the
     # devel version. Please note that the development snapshot in Debian does
@@ -194,7 +196,7 @@ def get_llvm_tool_names(tool: str) -> typing.List[str]:
         names.append(tool + suffix)
     return names
 
-def detect_scanbuild() -> typing.List[str]:
+def detect_scanbuild() -> T.List[str]:
     """ Look for scan-build binary on build platform
 
     First, if a SCANBUILD env variable has been provided, give it precedence
@@ -225,7 +227,7 @@ def detect_scanbuild() -> typing.List[str]:
             return [tool]
     return []
 
-def detect_clangformat() -> typing.List[str]:
+def detect_clangformat() -> T.List[str]:
     """ Look for clang-format binary on build platform
 
     Do the same thing as detect_scanbuild to find clang-format except it
@@ -397,7 +399,7 @@ def detect_msys2_arch():
         return os.environ['MSYSTEM_CARCH']
     return None
 
-def detect_machine_info(compilers: typing.Optional[CompilersDict] = None) -> MachineInfo:
+def detect_machine_info(compilers: T.Optional[CompilersDict] = None) -> MachineInfo:
     """Detect the machine we're running on
 
     If compilers are not provided, we cannot know as much. None out those
@@ -585,11 +587,19 @@ class Environment:
             self.default_objcpp = ['c++', 'g++']
             self.default_cs = ['csc', 'mcs']
         else:
-            self.default_c = ['cc', 'gcc', 'clang', 'pgcc', 'icc']
-            self.default_cpp = ['c++', 'g++', 'clang++', 'pgc++', 'icpc']
+            if platform.machine().lower() == 'e2k':
+                # There are no objc or objc++ compilers for Elbrus,
+                # and there's no clang which can build binaries for host.
+                self.default_c = ['cc', 'gcc', 'lcc']
+                self.default_cpp = ['c++', 'g++', 'l++']
+                self.default_objc = []
+                self.default_objcpp = []
+            else:
+                self.default_c = ['cc', 'gcc', 'clang', 'pgcc', 'icc']
+                self.default_cpp = ['c++', 'g++', 'clang++', 'pgc++', 'icpc']
+                self.default_objc = ['cc', 'gcc', 'clang']
+                self.default_objcpp = ['c++', 'g++', 'clang++']
             self.default_fortran = ['gfortran', 'flang', 'pgfortran', 'ifort', 'g95']
-            self.default_objc = ['cc', 'gcc', 'clang']
-            self.default_objcpp = ['c++', 'g++', 'clang++']
             self.default_cs = ['mcs', 'csc']
         self.default_d = ['ldc2', 'ldc', 'gdc', 'dmd']
         self.default_java = ['javac']
@@ -732,7 +742,7 @@ class Environment:
                 errmsg += '\nRunning "{0}" gave "{1}"'.format(c, e)
         raise EnvironmentException(errmsg)
 
-    def _guess_win_linker(self, compiler: typing.List[str], comp_class: Compiler,
+    def _guess_win_linker(self, compiler: T.List[str], comp_class: Compiler,
                           for_machine: MachineChoice, *,
                           use_linker_prefix: bool = True) -> 'DynamicLinker':
         self.coredata.add_lang_args(comp_class.language, comp_class, for_machine, self)
@@ -747,8 +757,8 @@ class Environment:
 
         check_args += self.coredata.compiler_options[for_machine][comp_class.language + '_args'].value
 
-        override = []  # type: typing.List[str]
-        value = self.binaries[for_machine].lookup_entry('ld')
+        override = []  # type: T.List[str]
+        value = self.binaries[for_machine].lookup_entry(comp_class.language + '_ld')
         if value is not None:
             override = comp_class.use_linker_args(value[0])
             check_args += override
@@ -757,7 +767,7 @@ class Environment:
         if o.startswith('LLD'):
             if '(compatible with GNU linkers)' in o:
                 return LLVMDynamicLinker(
-                    compiler, for_machine, 'lld', comp_class.LINKER_PREFIX,
+                    compiler, for_machine, comp_class.LINKER_PREFIX,
                     override, version=search_version(o))
 
         if value is not None:
@@ -791,9 +801,9 @@ class Environment:
                 "%PATH% variable to resolve this.")
         raise EnvironmentException('Unable to determine dynamic linker')
 
-    def _guess_nix_linker(self, compiler: typing.List[str], comp_class: typing.Type[Compiler],
+    def _guess_nix_linker(self, compiler: T.List[str], comp_class: T.Type[Compiler],
                           for_machine: MachineChoice, *,
-                          extra_args: typing.Optional[typing.List[str]] = None) -> 'DynamicLinker':
+                          extra_args: T.Optional[T.List[str]] = None) -> 'DynamicLinker':
         """Helper for guessing what linker to use on Unix-Like OSes.
 
         :compiler: Invocation to use to get linker
@@ -802,7 +812,7 @@ class Environment:
         :extra_args: Any additional arguments required (such as a source file)
         """
         self.coredata.add_lang_args(comp_class.language, comp_class, for_machine, self)
-        extra_args = typing.cast(typing.List[str], extra_args or [])
+        extra_args = T.cast(T.List[str], extra_args or [])
         extra_args += self.coredata.compiler_options[for_machine][comp_class.language + '_args'].value
 
         if isinstance(comp_class.LINKER_PREFIX, str):
@@ -810,8 +820,8 @@ class Environment:
         else:
             check_args = comp_class.LINKER_PREFIX + ['--version'] + extra_args
 
-        override = []  # type: typing.List[str]
-        value = self.binaries[for_machine].lookup_entry('ld')
+        override = []  # type: T.List[str]
+        value = self.binaries[for_machine].lookup_entry(comp_class.language + '_ld')
         if value is not None:
             override = comp_class.use_linker_args(value[0])
             check_args += override
@@ -820,7 +830,7 @@ class Environment:
         v = search_version(o)
         if o.startswith('LLD'):
             linker = LLVMDynamicLinker(
-                compiler, for_machine, 'lld', comp_class.LINKER_PREFIX, override, version=v)  # type: DynamicLinker
+                compiler, for_machine, comp_class.LINKER_PREFIX, override, version=v)  # type: DynamicLinker
         elif e.startswith('lld-link: '):
             # Toolchain wrapper got in the way; this happens with e.g. https://github.com/mstorsjo/llvm-mingw
             # Let's try to extract the linker invocation command to grab the version.
@@ -836,30 +846,29 @@ class Environment:
                 _, o, e = Popen_safe([linker_cmd, '--version'])
                 v = search_version(o)
 
-            linker = LLVMDynamicLinker(compiler, for_machine, 'lld', comp_class.LINKER_PREFIX, override, version=v)
+            linker = LLVMDynamicLinker(compiler, for_machine, comp_class.LINKER_PREFIX, override, version=v)
         # first is for apple clang, second is for real gcc, the third is icc
         elif e.endswith('(use -v to see invocation)\n') or 'macosx_version' in e or 'ld: unknown option:' in e:
             if isinstance(comp_class.LINKER_PREFIX, str):
                 _, _, e = Popen_safe(compiler + [comp_class.LINKER_PREFIX + '-v'] + extra_args)
             else:
                 _, _, e = Popen_safe(compiler + comp_class.LINKER_PREFIX + ['-v'] + extra_args)
-            i = 'APPLE ld'
             for line in e.split('\n'):
                 if 'PROJECT:ld' in line:
                     v = line.split('-')[1]
                     break
             else:
                 v = 'unknown version'
-            linker = AppleDynamicLinker(compiler, for_machine, i, comp_class.LINKER_PREFIX, override, version=v)
+            linker = AppleDynamicLinker(compiler, for_machine, comp_class.LINKER_PREFIX, override, version=v)
         elif 'GNU' in o:
             if 'gold' in o:
-                i = 'GNU ld.gold'
+                cls = GnuGoldDynamicLinker
             else:
-                i = 'GNU ld.bfd'
-            linker = GnuDynamicLinker(compiler, for_machine, i, comp_class.LINKER_PREFIX, override, version=v)
+                cls = GnuBFDDynamicLinker
+            linker = cls(compiler, for_machine, comp_class.LINKER_PREFIX, override, version=v)
         elif 'Solaris' in e or 'Solaris' in o:
             linker = SolarisDynamicLinker(
-                compiler, for_machine, 'solaris', comp_class.LINKER_PREFIX, override,
+                compiler, for_machine, comp_class.LINKER_PREFIX, override,
                 version=search_version(e))
         else:
             raise EnvironmentException('Unable to determine dynamic linker')
@@ -938,6 +947,7 @@ class Environment:
                     cls = GnuCCompiler if lang == 'c' else GnuCPPCompiler
 
                 linker = self._guess_nix_linker(compiler, cls, for_machine)
+
                 return cls(
                     ccache + compiler, version, for_machine, is_cross,
                     info, exe_wrap, defines, full_version=full_version,
@@ -945,6 +955,7 @@ class Environment:
 
             if 'Emscripten' in out:
                 cls = EmscriptenCCompiler if lang == 'c' else EmscriptenCPPCompiler
+                self.coredata.add_lang_args(cls.language, cls, for_machine, self)
                 return cls(
                     ccache + compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version)
@@ -965,6 +976,7 @@ class Environment:
                 full_version = arm_ver_str
                 cls = ArmclangCCompiler if lang == 'c' else ArmclangCPPCompiler
                 linker = ArmClangDynamicLinker(for_machine, version=version)
+                self.coredata.add_lang_args(cls.language, cls, for_machine, self)
                 return cls(
                     ccache + compiler, version, for_machine, is_cross, info,
                     exe_wrap, full_version=full_version, linker=linker)
@@ -1048,7 +1060,7 @@ class Environment:
             if 'PGI Compilers' in out:
                 cls = PGICCompiler if lang == 'c' else PGICPPCompiler
                 self.coredata.add_lang_args(cls.language, cls, for_machine, self)
-                linker = PGIDynamicLinker(compiler, for_machine, 'pgi', cls.LINKER_PREFIX, [], version=version)
+                linker = PGIDynamicLinker(compiler, for_machine, cls.LINKER_PREFIX, [], version=version)
                 return cls(
                     ccache + compiler, version, for_machine, is_cross,
                     info, exe_wrap, linker=linker)
@@ -1060,12 +1072,14 @@ class Environment:
                     exe_wrap, full_version=full_version, linker=l)
             if 'ARM' in out:
                 cls = ArmCCompiler if lang == 'c' else ArmCPPCompiler
+                self.coredata.add_lang_args(cls.language, cls, for_machine, self)
                 linker = ArmDynamicLinker(for_machine, version=version)
                 return cls(
                     ccache + compiler, version, for_machine, is_cross,
                     info, exe_wrap, full_version=full_version, linker=linker)
             if 'RX Family' in out:
                 cls = CcrxCCompiler if lang == 'c' else CcrxCPPCompiler
+                self.coredata.add_lang_args(cls.language, cls, for_machine, self)
                 linker = CcrxDynamicLinker(for_machine, version=version)
                 return cls(
                     ccache + compiler, version, for_machine, is_cross, info,
@@ -1114,7 +1128,7 @@ class Environment:
             cpp_compiler = self.detect_cpp_compiler(for_machine)
             cls = CudaCompiler
             self.coredata.add_lang_args(cls.language, cls, for_machine, self)
-            linker = CudaLinker(compiler, for_machine, 'nvlink', CudaCompiler.LINKER_PREFIX, [], version=CudaLinker.parse_version())
+            linker = CudaLinker(compiler, for_machine, CudaCompiler.LINKER_PREFIX, [], version=CudaLinker.parse_version())
             return cls(ccache + compiler, version, for_machine, is_cross, exe_wrap, host_compiler=cpp_compiler, info=info, linker=linker)
         raise EnvironmentException('Could not find suitable CUDA compiler: "' + ' '.join(compilers) + '"')
 
@@ -1199,7 +1213,7 @@ class Environment:
                 if 'PGI Compilers' in out:
                     cls = PGIFortranCompiler
                     self.coredata.add_lang_args(cls.language, cls, for_machine, self)
-                    linker = PGIDynamicLinker(compiler, for_machine, 'pgi',
+                    linker = PGIDynamicLinker(compiler, for_machine,
                                               cls.LINKER_PREFIX, [], version=version)
                     return cls(
                         compiler, version, for_machine, is_cross, info, exe_wrap,
@@ -1253,7 +1267,7 @@ class Environment:
                 popen_exceptions[' '.join(compiler + arg)] = e
                 continue
             version = search_version(out)
-            if 'Free Software Foundation' in out or ('e2k' in out and 'lcc' in out):
+            if 'Free Software Foundation' in out:
                 defines = self.get_gnu_compiler_defines(compiler)
                 if not defines:
                     popen_exceptions[' '.join(compiler)] = 'no pre-processor defines'
@@ -1354,6 +1368,7 @@ class Environment:
 
         cc = self.detect_c_compiler(for_machine)
         is_link_exe = isinstance(cc.linker, VisualStudioLikeLinkerMixin)
+        override = self.binaries[for_machine].lookup_entry('rust_ld')
 
         for compiler in compilers:
             if isinstance(compiler, str):
@@ -1377,33 +1392,47 @@ class Environment:
                 # the default use that, and second add the necessary arguments
                 # to rust to use -fuse-ld
 
-                extra_args = {}
-                always_args = []
-                if is_link_exe:
-                    compiler.extend(['-C', 'linker={}'.format(cc.linker.exelist[0])])
-                    extra_args['direct'] = True
-                    extra_args['machine'] = cc.linker.machine
-                elif not ((info.is_darwin() and isinstance(cc, AppleClangCCompiler)) or
-                          isinstance(cc, GnuCCompiler)):
+                if override is None:
+                    extra_args = {}
+                    always_args = []
+                    if is_link_exe:
+                        compiler.extend(['-C', 'linker={}'.format(cc.linker.exelist[0])])
+                        extra_args['direct'] = True
+                        extra_args['machine'] = cc.linker.machine
+                    elif not ((info.is_darwin() and isinstance(cc, AppleClangCCompiler)) or
+                              isinstance(cc, GnuCCompiler)):
+                        c = cc.exelist[1] if cc.exelist[0].endswith('ccache') else cc.exelist[0]
+                        compiler.extend(['-C', 'linker={}'.format(c)])
+
+                    # This trickery with type() gets us the class of the linker
+                    # so we can initialize a new copy for the Rust Compiler
+                    if is_link_exe:
+                        linker = type(cc.linker)(for_machine, always_args, exelist=cc.linker.exelist,
+                                                 version=cc.linker.version, **extra_args)
+                    else:
+                        linker = type(cc.linker)(compiler, for_machine, cc.LINKER_PREFIX,
+                                                 always_args=always_args, version=cc.linker.version,
+                                                 **extra_args)
+                elif 'link' in override[0]:
+                    linker = self._guess_win_linker(
+                        override, RustCompiler, for_machine, use_linker_prefix=False)
+                    linker.direct = True
+                else:
+                    # We're creating a new type of "C" compiler, that has rust
+                    # as it's language. This is gross, but I can't figure out
+                    # another way to handle this, because rustc is actually
+                    # invoking the c compiler as it's linker.
+                    b = type('b', (type(cc), ), {})
+                    b.language = RustCompiler.language
+                    linker = self._guess_nix_linker(cc.exelist, b, for_machine)
+
+                    # Of course, we're not going to use any of that, we just
+                    # need it to get the proper arguments to pass to rustc
                     c = cc.exelist[1] if cc.exelist[0].endswith('ccache') else cc.exelist[0]
                     compiler.extend(['-C', 'linker={}'.format(c)])
+                    compiler.extend(['-C', 'link-args={}'.format(' '.join(cc.use_linker_args(override[0])))])
 
-                value = self.binaries[for_machine].lookup_entry('ld')
-                if value is not None:
-                    for a in cc.use_linker_args(value[0]):
-                        always_args.extend(['-C', 'link-arg={}'.format(a)])
-
-                # This trickery with type() gets us the class of the linker
-                # so we can initialize a new copy for the Rust Compiler
-
-                if is_link_exe:
-                    linker = type(cc.linker)(for_machine, always_args, exelist=cc.linker.exelist,
-                                             version=cc.linker.version, **extra_args)
-                else:
-                    linker = type(cc.linker)(compiler, for_machine, cc.linker.id, cc.LINKER_PREFIX,
-                                             always_args=always_args, version=cc.linker.version,
-                                             **extra_args)
-
+                self.coredata.add_lang_args(RustCompiler.language, RustCompiler, for_machine, self)
                 return RustCompiler(
                     compiler, version, for_machine, is_cross, info, exe_wrap,
                     linker=linker)
